# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain
   :platform: Unix, Windows
   :synopsis: Canonical CtsTextMetadata Service and Canonical URN tool suite for Python

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

__version__ = "3.0.1"
