# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.common.reference
   :synopsis: URN related objects

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>

"""
from ._base import NodeId, BaseCitationSet, BaseReference, BaseReferenceSet
from ._capitains_cts import Citation, CtsReference, CtsReferenceSet, URN
from ._dts_1 import DtsCitation, DtsCitationSet, DtsReference, DtsReferenceSet
